#include <lektor/common.h>
#include <lektor/net.h>
#include <lektor/cmd.h>
#include <lektor/config.h>
#include <lektor/segv.h>

#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <netdb.h>
#include <signal.h>

#define LKT_KEY_VALUE_SEP ": \n\t\0"

#define HANDLE_STATUS(buffer, ok)            \
    if (STR_NMATCH(buffer, "OK", 2)) {       \
        ok;                                  \
    } else if (STR_NMATCH(buffer, "ACK", 3)) \
        LOG_FATAL("Got the 'ACK' status from lektord: %s", buffer);

/* Type definition. */

#define ROW_FORMAT_ONELINE "oneline"
#define ROW_FORMAT_PRETTY  "pretty"
#define ROW_FORMAT_AUTO    "auto"
typedef void (*row_printer)(char *);
row_printer row_print = NULL;

/* Global variables. */

#define LKT_MAX_OPTIONS 4 /* Number of options that can be passed to lkt before the commands. */
static const char *host;
static const char *port;
static const char *pwd;
static const char *row_format;

/* Default queue length to query */
static const char *LKT_QUEUE_DEFAULT[] = { "10" };

/* Small screen column count, to autoset the 'pretty' printer. The standard
 * 80x25 terminal is small! */
#define LKT_SMALL_SCREEN_THRESHOLD 82
#define IS_SMALL_SCREEN            (___screen_detected && (___column_count <= LKT_SMALL_SCREEN_THRESHOLD))
static unsigned ___column_count = 0;
static bool ___screen_detected  = false;
CONSTRUCTOR_FUNCTION
setup_column_count(void)
{
    struct winsize size;
    if (0 == ioctl(STDOUT_FILENO, TIOCGWINSZ, &size)) {
        ___screen_detected = true;
        ___column_count    = size.ws_col;
    }
}

/* Row printer, should be queue/playlist/search agnostic, i.e. should handle
 * the following lines in a similar way (be aware of the space):
 * - 7452 vo - multi / Code Geass - IS1 - Continued Story [Kubat]
 * -  90: 7452 vo - multi / Code Geass - IS1 - Continued Story [Kubat]
 * - 5: 7452 vo - multi / Code Geass - IS1 - Continued Story [Kubat]
 *
 * NOTE: If the line is invalid, it shall only display it. This is done for
 * generecity in '___continuation_calls' functions and such.
 */

PRIVATE_FUNCTION void
___row_printer_online(char *line)
{
    fprintf(stdout, "%s", line);
}

PRIVATE_FUNCTION void
___row_printer_pretty(char *line)
{
    line += strspn(line, " \t");
    char *last = &line[strlen(line) - 1];
    FAIL_IF(last == NULL, "Invalid line, might be empty")
    if (*last == '\n') {
        *last = '\0';
        last--;
    }

    /* The separator between the 'cat - language' and the source title, etc. */
    char *title = strstr(line, " / ");

    /* Invalid line, a kara line must end with the: [author]
     * It's ok, the other way this function is called is with the stickers and
     * playlists searches, which must not contain the ']' character.
     * If the separator is not found, the line is greatly invalid, but print it
     * anyway (not our problem). */
    if (*last != ']' || title == NULL)
        fprintf(stdout, "%s\n", line);

    /* Get the title */
    *title = '\0';
    title += strlen(" / ");

    /* Get the author */
    char *author = strstr(title, " [");
    *author      = '\0';
    author += strlen(" [");
    FAIL_IF(author == NULL, "Invalid row, faild to find the author");
    last = strstr(author, "]");
    FAIL_IF(last == NULL, "Invalid author string, missing ']'");
    *last = '\0';

    /* Print it! */
    fprintf(stdout, "%s by %s\n%s\n\n", line, author, title);
}

PRIVATE_FUNCTION void
setup_row_printer(void)
{
    if (NULL == row_format || STR_MATCH(row_format, ROW_FORMAT_AUTO))
        row_print = IS_SMALL_SCREEN ? ___row_printer_pretty : ___row_printer_online;

    else if (NULL == row_format || STR_MATCH(row_format, ROW_FORMAT_ONELINE))
        row_print = ___row_printer_online;

    else if (STR_MATCH(row_format, ROW_FORMAT_PRETTY))
        row_print = ___row_printer_pretty;

    else
        LOG_FATAL("Invalid unknown row printer format: %s", row_format);
}

/* Communication functions and fonction that interact with lektor stuff. */

PRIVATE_FUNCTION void
concat_strings(char *const res, size_t len, const char **tab, size_t count)
{
    size_t i;
    for (i = 0; i < count - 1; ++i) {
        strncat(res, tab[i], len - 1);
        strncat(res, " ", len - 1);
    }
    strncat(res, tab[i], len - 1);
}

PRIVATE_FUNCTION int
lkt_get_query_type(struct cmd_args *args, char *const regex, size_t regex_len)
{
    if (args->argc == 0) {
        LOG_ERROR("LKT", "Invalid empty query");
        return 1;
    }

    /* id, type, lang, category */
    if (args->argc == 1) {
        /* Type ?== id */
        errno   = 0;
        long id = strtol(args->argv[0], NULL, 10);
        if (id != 0 && errno == 0) {
            safe_snprintf(regex, regex_len, "id://%ld", id);
            return 0;
        }

        /* Type ?== type OR Type ?== category OR Type ?== lang */
#define kara_type(tp)                                                \
    if (STR_MATCH(tp, args->argv[0])) {                              \
        safe_snprintf(regex, regex_len, "type://%s", args->argv[0]); \
        return 0;                                                    \
    }
#define kara_category(ct)                                                \
    if (STR_MATCH(ct, args->argv[0])) {                                  \
        safe_snprintf(regex, regex_len, "category://%s", args->argv[0]); \
        return 0;                                                        \
    }
#define kara_language(lg)                                            \
    if (STR_MATCH(lg, args->argv[0])) {                              \
        safe_snprintf(regex, regex_len, "lang://%s", args->argv[0]); \
        return 0;                                                    \
    }
#include <lektor/internal/database.def>
#undef kara_category
#undef kara_language
#undef kara_type
    }

    memset(regex, 0, regex_len * sizeof(char));

    /* Mendatory author, playlist/plt. Explicit type for lang, category and type. */
    if (args->argc >= 2) {
#define ___explicit_with_concat_strings(matches, uri)                               \
    if (STR_MATCH(matches, args->argv[0])) {                                        \
        strncat(regex, uri, regex_len - 1);                                         \
        concat_strings(regex, regex_len, args->argv + 1, (size_t)(args->argc - 1)); \
        return 0;                                                                   \
    }
#define ___explicit_with_snprintf(matches, uri)                   \
    if (STR_MATCH(matches, args->argv[0])) {                      \
        safe_snprintf(regex, regex_len, uri "%s", args->argv[1]); \
        return 0;                                                 \
    }

        // clang-format off
        ___explicit_with_concat_strings("author",   "author://");   /* Type ?== author   */
        ___explicit_with_concat_strings("plt",      "playlist://"); /* Type ?== playlist */
        ___explicit_with_concat_strings("playlist", "playlist://"); /* Type ?== playlist */
        ___explicit_with_snprintf("cat",            "category://"); /* Type ?== category */
        ___explicit_with_snprintf("category",       "category://"); /* Type ?== category */
        ___explicit_with_snprintf("lang",           "lang://");     /* Type ?== lang     */
        ___explicit_with_snprintf("language",       "lang://");     /* Type ?== lang     */
        ___explicit_with_snprintf("type",           "type://");     /* Type ?== type     */
        // clang-format on

#undef ___explicit_with_snprintf
#undef ___explicit_concat_strings
    }

    /* If 'query' is specified, skip it */
    int index = STR_MATCH("query", args->argv[0]) ? 1 : 0;
    strncat(regex, "query://%", regex_len - 1);
    concat_strings(regex, regex_len, args->argv + index, (size_t)(args->argc - index));
    strncat(regex, "%", regex_len - 1);
    return 0;
}

PRIVATE_FUNCTION int
read_socket(FILE *sock, char *buff, size_t max_len)
{
    size_t i, len = 0;
    memset(buff, 0, max_len * sizeof(char));
    for (i = 0; i < max_len - 1; ++i) {
        len = fread(buff + i, sizeof(char), 1, sock);
        if (buff[i] == '\n' || len != 1)
            break;
    }
    FAIL_IF(len == 0, "Connexion error");
    return (int)i;
}

PRIVATE_FUNCTION EXIT_FUNCTION
exit_with_status(FILE *sock, char *buff)
{
    read_socket(sock, buff, 2);
    FAIL_IF(buff[0] != 'O' && buff[1] != 'K', "ACK");
    exit(EXIT_SUCCESS);
}

PRIVATE_FUNCTION FILE *
create_socket(const char *host_arg, const char *port_arg)
{
    FILE *socket_desc;
    int cx;
    int domain;
    struct sockaddr *ptr_sock;
    socklen_t sock_len;

    /* Unix socket */
    if (STR_NMATCH(port_arg, "unix", 4)) {
        struct sockaddr_un sock;

        memset(&sock, 0, sizeof(struct sockaddr_un));
        sock.sun_family = AF_UNIX;
        strncpy(sock.sun_path, host_arg, sizeof(sock.sun_path) - 1);

        ptr_sock = (struct sockaddr *)&sock;
        domain   = AF_UNIX;
        sock_len = sizeof(struct sockaddr_un);
    }

    /* TCP socket */
    else {
        struct addrinfo hints, *found;
        struct sockaddr sock;

        memset(&hints, 0, sizeof(struct addrinfo));
        hints.ai_flags    = 0;
        hints.ai_family   = AF_INET;
        hints.ai_socktype = SOCK_STREAM;

        if (getaddrinfo(host_arg, port_arg, &hints, &found) != 0 || (found == NULL))
            LOG_FATAL("getaddrinfo failed, this can be due to php "
                      "invocation of lkt or an invalid hostname...");

        sock     = *found->ai_addr;
        sock_len = found->ai_addrlen;
        freeaddrinfo(found);

        ptr_sock = &sock;
        domain   = AF_INET;
    }

    /* Common part */
    cx = socket(domain, SOCK_STREAM, 0);
    FAIL_IF(cx <= 0 || connect(cx, ptr_sock, sock_len),
            "Connect to socket failed, is lektord launched?");
    socket_desc = fdopen(cx, "r+");
    FAIL_IF(!socket_desc, "Failed to connect to lektord");
    return socket_desc;
}

PRIVATE_FUNCTION FILE *
lkt_connect(void)
{
    char buff[LKT_MESSAGE_MAX];
    FILE *sock = create_socket(host, port);
    read_socket(sock, buff, LKT_MESSAGE_MAX);
    return sock;
}

PRIVATE_FUNCTION void
write_socket(FILE *sock, const char *format, ...)
{
    char buff[LKT_MESSAGE_MAX];
    va_list ap;

    va_start(ap, format);
    int size = safe_vsnprintf(buff, LKT_MESSAGE_MAX - 1, format, ap);

    FAIL_IF(size < 0, "Failed to write to temporary buffer");
    FAIL_IF(fwrite(buff, sizeof(char), (size_t)size, sock) < (unsigned int)size, "Connexion error");

    va_end(ap);
}

PRIVATE_FUNCTION EXIT_FUNCTION
lkt_send_and_exit(const char *buffer, size_t len)
{
    write_socket(lkt_connect(), buffer, len);
    exit(EXIT_SUCCESS);
}

PRIVATE_FUNCTION char *
lkt_skip_key(char *buffer)
{
    size_t len = strcspn(buffer, LKT_KEY_VALUE_SEP);
    char *ret  = &buffer[len + 1];
    len        = strspn(ret, ": ");
    return ret + len;
}

PRIVATE_FUNCTION void
send_cmd_with_uri(FILE *sock, char *cmd, int argc, const char **argv)
{
    char buf[LKT_MESSAGE_MAX];
    struct cmd_args args = CMD_ARGS_FROM(argc, argv);

    memset(buf, 0, LKT_MESSAGE_MAX * sizeof(char));
    FAIL_IF(lkt_get_query_type(&args, buf, LKT_MESSAGE_MAX), "Query is invalid");
    write_socket(sock, "%s %s\n", cmd, buf);
}

/* Functions implementing options. */

#define just_send_two_args(func, cmd)                                     \
    PRIVATE_FUNCTION EXIT_FUNCTION func(struct cmd_args *args)            \
    {                                                                     \
        FAIL_IF(args->argc != 2, "Need two arguments");                   \
        FILE *sock = lkt_connect();                                       \
        char buff[2];                                                     \
        write_socket(sock, cmd " %s %s\n", args->argv[0], args->argv[1]); \
        exit_with_status(sock, buff);                                     \
    }
// clang-format off
just_send_two_args(queue_swap__,   "swap");
just_send_two_args(queue_swapid__, "swapid");
// clang-format on
#undef just_send_two_args

#define just_send_one_arg(func, cmd)                           \
    PRIVATE_FUNCTION EXIT_FUNCTION func(struct cmd_args *args) \
    {                                                          \
        FAIL_IF(args->argc != 1, "Invalid argument");          \
        FILE *sock = lkt_connect();                            \
        char buff[2];                                          \
        write_socket(sock, cmd " %s\n", args->argv[0]);        \
        exit_with_status(sock, buff);                          \
    }
// clang-format off
just_send_one_arg(stickers_create__,  "sticker __create");
just_send_one_arg(stickers_destroy__, "sticker __destroy");
just_send_one_arg(plt_destroy__,      "rm");
just_send_one_arg(plt_create__,       "playlistadd");
just_send_one_arg(queue_dump__,       "__dump");
// clang-format on
#undef just_send_one_arg

#define just_send(func, msg) /* Just send a simple string functions */ \
    PRIVATE_FUNCTION EXIT_FUNCTION func(struct cmd_args *args)         \
    {                                                                  \
        FAIL_IF(args->argc, "This command takes no arguments");        \
        lkt_send_and_exit(msg, sizeof(msg));                           \
    }
// clang-format off
just_send(queue_clear__,   "clear\n");
just_send(queue_crop__,    "crop\n");
just_send(next__,          "next\n");
just_send(prev__,          "previous\n");
just_send(stop__,          "stop\n");
just_send(pause__,         "pause 1\n");
just_send(unpause__,       "pause 0\n");
just_send(shuffle__,       "shuffle\npause 0\n");
just_send(queue_flatten__, "__flat\n");
// clang-format on
#undef just_send

PRIVATE_FUNCTION EXIT_FUNCTION
simple_send_with_password__(const char *cmd)
{
    FAIL_UNLESS(pwd, "Password not provided");
    FILE *sock = lkt_connect();
    write_socket(sock, "command_list_begin\n");
    write_socket(sock, "password %s\n", pwd);
    write_socket(sock, "%s\n", cmd);
    write_socket(sock, "command_list_end\n");
    exit(EXIT_SUCCESS);
}

PRIVATE_FUNCTION EXIT_FUNCTION
restart__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 0, "Invalid argument, the previous command takes no arguments");
    simple_send_with_password__("__restart");
}

PRIVATE_FUNCTION EXIT_FUNCTION
kill__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 0, "Invalid argument, the previous command takes no arguments");
    simple_send_with_password__("kill");
}

PRIVATE_FUNCTION EXIT_FUNCTION
rescan_or_update__(struct cmd_args *args, const char *cmd)
{
    FAIL_UNLESS(pwd, "Password not provided");
    FILE *sock = lkt_connect();

    /* All the db */
    if (args->argc == 0) {
        LOG_INFO("LKT-ADM", "Send '%s' without query", cmd);
        write_socket(sock, "command_list_begin\n");
        write_socket(sock, "password %s\n", pwd);
        write_socket(sock, "%s\n", cmd);
        write_socket(sock, "command_list_end\n");
    }

    /* With a query */
    else {
        char buff[LKT_MESSAGE_MAX];
        char cmd_buff[LKT_MESSAGE_MAX];
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));

        FAIL_IF(lkt_get_query_type(args, buff, LKT_MESSAGE_MAX), "Query is not valid");
        safe_snprintf(cmd_buff, LKT_MESSAGE_MAX, "%s %s", cmd, buff);
        LOG_INFO("LKT-ADM", "Send '%s'", cmd_buff);

        write_socket(sock, "command_list_begin\n");
        write_socket(sock, "password %s\n", pwd);
        write_socket(sock, "%s\n", cmd_buff);
        write_socket(sock, "command_list_end\n");
    }

    exit(EXIT_SUCCESS);
}

#define ___just_send_with_password(func, cmd)                  \
    PRIVATE_FUNCTION EXIT_FUNCTION func(struct cmd_args *args) \
    {                                                          \
        rescan_or_update__(args, cmd);                         \
    }                                                          \
// clang-format off
___just_send_with_password(rescan__, "rescan");
___just_send_with_password(update__, "update");
___just_send_with_password(dry_update__, "__dry_update");
___just_send_with_password(import__, "__import");
___just_send_with_password(populate__, "__rescan");
// clang-format on
#undef ___just_send_with_password

PRIVATE_FUNCTION EXIT_FUNCTION
queue_replace__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 1, "Invalid argument");
    char buff[LKT_MESSAGE_MAX];
    bool play  = false;
    FILE *sock = lkt_connect();

    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);
        size_t len = strcspn(buff, LKT_KEY_VALUE_SEP);

        if (STR_NMATCH(buff, "state", len)) {
            char *it = lkt_skip_key(buff);
            play     = STR_NMATCH(it, "play", 4);
            break;
        }

        HANDLE_STATUS(buff, { break; });
    }

    fclose(sock);
    sock = lkt_connect();
    write_socket(sock, "command_list_begin\n");
    write_socket(sock, "clear\n");
    write_socket(sock, "add playlist://%s\n", args->argv[0]);
    if (play)
        write_socket(sock, "play\n");
    write_socket(sock, "command_list_end\n");
    exit_with_status(sock, buff);
}

PRIVATE_FUNCTION EXIT_FUNCTION
config__(struct cmd_args *args)
{
    int ret = EXIT_SUCCESS;
    if (args->argc == 0)
        fprintf(stdout, "%s", lkt_default_config_file);
    exit(ret);
}

PRIVATE_FUNCTION EXIT_FUNCTION
play__(struct cmd_args *args)
{
    FAIL_IF(args->argc > 1, "Invalid argument");

    int pos = (args->argc != 0) ? atoi(args->argv[0]) : 0;

    static const char status__[]        = "status\n";
    static const char cmd_play__[]      = "play\n";
    static const char cmd_play_from__[] = "play %d\n";
    static const char cmd_pause__[]     = "pause\n";

    char buff[LKT_MESSAGE_MAX];
    FILE *sock = lkt_connect();
    write_socket(sock, status__);

    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);
        size_t len = strcspn(buff, LKT_KEY_VALUE_SEP);

        if (STR_NMATCH(buff, "state", len)) {
            fclose(sock);

            if (STR_NMATCH(lkt_skip_key(buff), "stop", 4)) {
                if (!pos)
                    lkt_send_and_exit(cmd_play__, sizeof(cmd_play__)); /* In bytes. */
                else {
                    write_socket(lkt_connect(), cmd_play_from__, pos);
                    exit(EXIT_SUCCESS);
                }
            }

            else
                lkt_send_and_exit(cmd_pause__, sizeof(cmd_pause__)); /* In bytes. */

            LOG_UNREACHABLE;
        }
    }
}

PRIVATE_FUNCTION EXIT_FUNCTION
ping__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 0, "Invalid argument, the ping command takes no arguments");
    char buff[6];
    FILE *sock = lkt_connect();
    memset(buff, 0, sizeof(buff));
    write_socket(sock, "ping\n");
    read_socket(sock, buff, 6 * sizeof(char));
    FAIL_UNLESS(STR_NMATCH(buff, "OK", 2), "ACK");
    exit(write(1, "OK\n", sizeof("OK\n")) == 0);
}

PRIVATE_FUNCTION EXIT_FUNCTION
read_kara_and_exit(FILE *sock)
{
    char *mem = NULL;
    char buff[LKT_MESSAGE_MAX];
    lkt_assert(mem = calloc(6 * LKT_MESSAGE_MAX, sizeof(char)));
    lkt_assert(memset(mem, 0, 6 * LKT_MESSAGE_MAX * sizeof(char)));

    char *const title    = &mem[0];
    char *const author   = &mem[LKT_MESSAGE_MAX];
    char *const source   = &mem[2 * LKT_MESSAGE_MAX];
    char *const type     = &mem[3 * LKT_MESSAGE_MAX];
    char *const category = &mem[4 * LKT_MESSAGE_MAX];
    char *const language = &mem[5 * LKT_MESSAGE_MAX];

#define ___assign_mdt(what)              \
    if (STR_NMATCH(#what, buff, len)) {  \
        memcpy(what, value, value_size); \
    }

    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);

        const size_t len        = strcspn(buff, LKT_KEY_VALUE_SEP);
        char *value             = lkt_skip_key(buff);
        const size_t value_size = strcspn(value, "\n\0") * sizeof(char);

        ___assign_mdt(title);
        ___assign_mdt(author);
        ___assign_mdt(source);
        ___assign_mdt(type);
        ___assign_mdt(language);
        ___assign_mdt(category);

        /* At this point every key has been parsed. */
        HANDLE_STATUS(buff, { break; });
    }

#undef ___assign_mdt

    if (language[0])
        printf("%s - %s / %s - %s - %s [%s]\n", category, language, source, type, title, author);
    else if (category[0])
        printf("%s / %s - %s - %s [%s]\n", category, source, type, title, author);
    else
        abort();
    exit(EXIT_SUCCESS);
}

PRIVATE_FUNCTION EXIT_FUNCTION
queue_id__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 1, "Invalid argument, the current command takes one argument");
    char probe_id__[LKT_LINE_MAX];
    safe_snprintf(probe_id__, LKT_LINE_MAX, "playlistid %ld\n", strtol(args->argv[0], NULL, 0));
    FILE *sock = lkt_connect();
    write_socket(sock, probe_id__);
    read_kara_and_exit(sock);
}

PRIVATE_FUNCTION EXIT_FUNCTION
current__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 0, "Invalid argument, the current command takes no arguments");
    static const char current_song__[] = "currentsong\n";
    FILE *sock                         = lkt_connect();
    write_socket(sock, current_song__);
    read_kara_and_exit(sock);
}

PRIVATE_FUNCTION EXIT_FUNCTION
queue_pop__(struct cmd_args *args)
{
    FAIL_IF(args->argc, "Invalid argument");
    int song = 0;
    char buff[LKT_MESSAGE_MAX];
    FILE *sock = lkt_connect();

    /* Get lektor's status. */
    write_socket(sock, "status\n");

#define assign_int(str, var)              \
    if (STR_NMATCH(buff, str, len)) {     \
        var = (atoi(lkt_skip_key(buff))); \
        continue;                         \
    }
    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);
        size_t len = strcspn(buff, LKT_KEY_VALUE_SEP);
        assign_int("song", song);

        /* At this point every key has been parsed. */
        HANDLE_STATUS(buff, { break; });
    }
#undef assign_int
    song += 1; /* Needs the +1, see status command */

    fclose(sock);
    sock = lkt_connect();
    if (!song)
        exit(EXIT_FAILURE);
    write_socket(sock, "next\ndelete %d\n", song);
    exit_with_status(sock, buff);
}

PRIVATE_FUNCTION EXIT_FUNCTION
status__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 0, "Invalid argument, the status command takes no arguments");

    static const char *const status_str__ = "status\n";
    static const char *const stats_str__  = "stats\n";

    char buff[LKT_MESSAGE_MAX];
    char flags[24];

    bool play        = false;
    bool stopped     = true;
    bool is_updating = false;

    int time_pos      = 0;
    int time_duration = 0;
    int song_index    = 0;
    int plt_len       = 0;
    int it            = 0;
    int update_count  = 0;
    int update_tick   = 0;
    time_t update_ts  = 0;
    size_t len;

    memset(flags, 0, 24 * sizeof(char));
    FILE *sock = lkt_connect();

#define assign_flag(str, f)                                              \
    if (STR_NMATCH(buff, str, len) && (atoi(lkt_skip_key(buff)) == 1)) { \
        flags[it++] = f;                                                 \
        continue;                                                        \
    }
#define assign_two_int(str, var1, var2)                               \
    if (STR_NMATCH(buff, str, len)) {                                 \
        char *endptr = NULL;                                          \
        var1         = (int)(strtol(lkt_skip_key(buff), &endptr, 0)); \
        endptr += strspn(endptr, "-/:");                              \
        if (endptr)                                                   \
            var2 = (int)(strtol(endptr, NULL, 0));                    \
        continue;                                                     \
    }
#define assign_int(str, var)                              \
    if (STR_NMATCH(buff, str, len)) {                     \
        var = (int)(strtol(lkt_skip_key(buff), NULL, 0)); \
        continue;                                         \
    }

    /* Get lektor's status */
    write_socket(sock, status_str__);
    FOR_EVER
    {
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);
        len = strcspn(buff, LKT_KEY_VALUE_SEP);

        if (STR_NMATCH(buff, "state", len)) {
            char *itc = lkt_skip_key(buff);
            play      = STR_NMATCH(itc, "play", 4);
            stopped   = STR_NMATCH(itc, "stop", 4);
            continue;
        }

        assign_flag("random", 'r');
        assign_flag("repeat", 'l');
        assign_flag("single", 's');
        assign_flag("consume", 'c');
        assign_int("elapsed", time_pos);
        assign_int("duration", time_duration);
        assign_int("song", song_index);
        assign_int("playlistlength", plt_len);

        HANDLE_STATUS(buff, { break; });
    }

    /* Get lektor's stats */
    write_socket(sock, stats_str__);
    FOR_EVER
    {
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);
        len = strcspn(buff, LKT_KEY_VALUE_SEP);

        assign_int("__is_updating", is_updating) assign_int("db_update", update_ts);
        assign_two_int("__update_progress", update_tick, update_count);
        HANDLE_STATUS(buff, { break; });
    }

    /* End of communication */
    fclose(sock);

#undef assign_flag
#undef assign_int
#undef assign_two_int
#undef check_end

    struct tm *p_tm = localtime(&update_ts);
    len             = strftime(buff, LKT_MESSAGE_MAX - 1, "%F %H:%M:%S", p_tm);
    buff[len]       = '\0';

    int pla_m = time_pos / 60;
    int pla_s = time_pos % 60;
    int dur_m = time_duration / 60;
    int dur_s = time_duration % 60;

    printf("Lektor: %s\n"
           "Queue: #%d/%d\n"
           "Playback: %d:%02d/%d:%02d\n"
           "Flags: %s\n",
           play ? "play" : "stopped", song_index + 1, plt_len, pla_m, pla_s, dur_m, dur_s,
           flags[0] ? flags : "(none)");

    if (is_updating)
        printf("Updating: %d/%d%s\n", update_tick, update_count,
               update_count >= update_tick ? "" : " (invalid)");
    else
        printf("Last update: %s\n", buff);

    /* If there is a kara loaded in mpv, get this kara. */
    if (!stopped) {
        printf("Kara: ");
        current__(args);
    }

    exit(EXIT_SUCCESS);
}

PRIVATE_FUNCTION EXIT_FUNCTION
queue_remove__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 1, "Invalid argument, need onlt one argument: queue id <id>");

    static const char *cmd__ = "deleteid %d\n";
    int dumy                 = 0;
    FILE *sock               = lkt_connect();
    char buff[3];

    sscanf(args->argv[0], "%d", &dumy);
    if (dumy != 0) {
        write_socket(sock, cmd__, dumy);
        exit_with_status(sock, buff);
    }

    LOG_FATAL("Invalid argument");
}

PRIVATE_FUNCTION EXIT_FUNCTION
queue_delete__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 1, "Invalid argument, need onlt one argument: queue delete <pos>");

    static const char *cmd_id__ = "delete %d\n";
    int dumy                    = 0;
    FILE *sock                  = lkt_connect();
    char buff[3];

    sscanf(args->argv[0], "%d", &dumy);
    if (dumy != 0) {
        write_socket(sock, cmd_id__, dumy);
        exit_with_status(sock, buff);
    }

    LOG_FATAL("Invalid argument");
}

PRIVATE_FUNCTION EXIT_FUNCTION
queue_add__(struct cmd_args *args)
{
    char buff[3];
    FAIL_IF(args->argc < 1, "Invalid arguments");
    FILE *sock = lkt_connect();
    send_cmd_with_uri(sock, "add", args->argc, args->argv);
    exit_with_status(sock, buff);
}

PRIVATE_FUNCTION EXIT_FUNCTION
queue_insert__(struct cmd_args *args)
{
    char buff[3];
    FAIL_IF(args->argc < 1, "Invalid arguments");
    FILE *sock = lkt_connect();
    send_cmd_with_uri(sock, "__insert", args->argc, args->argv);
    exit_with_status(sock, buff);
}

PRIVATE_FUNCTION EXIT_FUNCTION
queue_seek__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 1, "The seek command needs one argument: queue seek <id>");

    char *endptr, buf[3];
    errno   = 0;
    long id = strtol(args->argv[0], &endptr, 0);

    FAIL_IF(STRTOL_ERROR(id) || (endptr == args->argv[0]), "Invalid argument, not an integer");
    FAIL_IF(*endptr != '\0', "Invalid argument, must be only one integer");

    FILE *sock = lkt_connect();
    write_socket(sock, "playid %ld\n", id);
    exit_with_status(sock, buf);
}

PRIVATE_FUNCTION EXIT_FUNCTION
queue_pos__(struct cmd_args *args)
{
    char buff[LKT_MESSAGE_MAX], *endptr;

    FAIL_IF(
        args->argc != 1,
        "Invalid argument for the pos command: queue pos <arg> where arg is a position or a range");

    long up           = 0;
    errno             = 0;
    long continuation = strtol(args->argv[0], &endptr, 0);

    FAIL_IF(STRTOL_ERROR(continuation) || (endptr == args->argv[0]),
            "Invalid argument, not an integer: queue pos <arg> where arg is a position or a range");

    if (*endptr != '\0') {
        /* A range */
        FAIL_IF(*(++endptr) == '\0', "Invalid argument, a range is two integers");
        up = atoi(endptr);
    }

    FILE *sock = NULL;
redo:
    sock = lkt_connect();
    if (up != 0)
        write_socket(sock, "playlist %d:%d\n", continuation, up);
    else
        write_socket(sock, "playlist %d\n", continuation);

    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);

        if (STR_NMATCH(buff, "continue:", strlen("continue:"))) {
            continuation = atoi(lkt_skip_key(buff));
            if (continuation > 0) {
                fclose(sock);
                goto redo;
            }
        }

        HANDLE_STATUS(buff, { exit(EXIT_SUCCESS); });
        row_print(buff);
    }
}

PRIVATE_FUNCTION EXIT_FUNCTION
queue_list__(struct cmd_args *args)
{
    char buff[LKT_MESSAGE_MAX], *endptr;
    long continuation = 0;
    long song_index   = 1;
    errno             = 0;

    /* Arguments stuff. */

    if (args->argc == 0)
        args->argv = LKT_QUEUE_DEFAULT;
    else
        FAIL_IF(args->argc != 1, "Invalid argument: queue <count?>");

    continuation = strtol(args->argv[0], &endptr, 0);
    FAIL_IF(STRTOL_ERROR(continuation) || (endptr == args->argv[0]),
            "Invalid argument, not an integer");
    FAIL_IF(*endptr != '\0', "Invalid argument");

    /* Get the current pos to get limits for the playlist command. */

    FILE *sock = lkt_connect();
    write_socket(sock, "status\n");

#define assign_int(str, var)              \
    if (STR_NMATCH(buff, str, len)) {     \
        var = (atoi(lkt_skip_key(buff))); \
        continue;                         \
    }
    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);
        size_t len = strcspn(buff, LKT_KEY_VALUE_SEP);
        assign_int("song", song_index);

        /* At this point every key has been parsed. */
        HANDLE_STATUS(buff, { break; });
    }
#undef assign_int
    fclose(sock);

    /* Get the content of the queue. */

    continuation = labs(continuation);
    song_index   = MAX(song_index + 1, 1);
    safe_snprintf(buff, LKT_MESSAGE_MAX, "%ld:%ld", song_index, song_index + continuation - 1);
    args->argc    = 1;
    args->argv[0] = buff;
    queue_pos__(args);
}

/* Functions implementing options, but for for playlists. */

PRIVATE_FUNCTION EXIT_FUNCTION
plt_add__(struct cmd_args *args)
{
    char buff[2];
    FAIL_IF(args->argc < 2,
            "Invalid argument, need at least three arguments: plt add <plt> <query>");

    char regex[LKT_MESSAGE_MAX];
    struct cmd_args args_regex = CMD_ARGS_NEXT(*args);

    FAIL_IF(lkt_get_query_type(&args_regex, regex, LKT_MESSAGE_MAX), "Query is invalid");
    FILE *sock = lkt_connect();
    write_socket(sock, "playlistadd %s %s\n", args->argv[0], regex);
    exit_with_status(sock, buff);
}

PRIVATE_FUNCTION EXIT_FUNCTION
plt_delete__(struct cmd_args *args)
{
    FILE *sock = lkt_connect();
    char buff[2];
    char cmd[128];
    FAIL_IF(args->argc < 3, "Invalid argument");
    snprintf(cmd, 128 - 1, "playlistdelete %s", args->argv[0]);
    cmd[127] = '\0';
    send_cmd_with_uri(sock, cmd, args->argc - 1, &args->argv[1]);
    exit_with_status(sock, buff);
}

PRIVATE_FUNCTION EXIT_FUNCTION
stickers_get__(struct cmd_args *args)
{
    char buff[LKT_MESSAGE_MAX];
    FILE *sock;

    if (args->argc == 2)
        write_socket(sock = lkt_connect(), "sticker get %s %s\n", args->argv[0], args->argv[1]);

    else if (args->argc == 3)
        write_socket(sock = lkt_connect(), "sticker get %s %s %s\n", args->argv[0], args->argv[1],
                     args->argv[2]);

    else if (args->argc == 5) {
        const char *op = args->argv[3];
        op             = op[0] == 'e' ? "=" : op[0] == 'l' ? "<" : op[0] == 'g' ? ">" : NULL;
        FAIL_IF(!op, "Invalid argument");
        write_socket(sock = lkt_connect(), "sticker get %s %s %s %s %s\n", args->argv[0],
                     args->argv[1], args->argv[2], op, args->argv[4]);
    }

    else
        LOG_FATAL("Invalid argument");

    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);
        HANDLE_STATUS(buff, { exit(EXIT_SUCCESS); });
        fprintf(stdout, "%s", buff);
    }
}

PRIVATE_FUNCTION EXIT_FUNCTION
stickers_set__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 4, "Invalid argument");
    FILE *sock = lkt_connect();
    char buff[2];
    write_socket(sock, "sticker set %s %s %s %s\n", args->argv[0], args->argv[1], args->argv[2],
                 args->argv[3]);
    exit_with_status(sock, buff);
}

PRIVATE_FUNCTION EXIT_FUNCTION
stickers_delete__(struct cmd_args *args)
{
    FILE *sock;
    char buff[2];
    if (args->argc == 2)
        write_socket(sock = lkt_connect(), "sticker delete %s %s", args->argv[0], args->argv[1]);
    else if (args->argc == 3)
        write_socket(sock = lkt_connect(), "sticker delete %s %s %s", args->argv[0], args->argv[1],
                     args->argv[2]);
    else
        LOG_FATAL("Invalid argument");
    exit_with_status(sock, buff);
}

/* Search functions. */

PRIVATE_FUNCTION EXIT_FUNCTION
___continuation_calls(const char *cmd)
{
    char buff[LKT_MESSAGE_MAX];
    int continuation = 0;
    FILE *sock       = NULL;

redo:
    sock = lkt_connect();

    write_socket(sock, "%d %s\n", continuation, cmd);
    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);

        if (STR_NMATCH(buff, "continue:", strlen("continue:"))) {
            continuation = atoi(lkt_skip_key(buff));
            fclose(sock);
            goto redo;
        }

        HANDLE_STATUS(buff, { exit(EXIT_SUCCESS); });

        if (STR_NMATCH(buff, "name: ", strlen("name: "))) {
            char *real_line = buff + sizeof("name: ") - sizeof(char);
            memmove(buff, real_line, (1 + strlen(real_line)) * sizeof(char));
        }

        row_print(buff);
    }
}

PRIVATE_FUNCTION EXIT_FUNCTION
list_plt__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 0, "Invalid argument number");
    ___continuation_calls("listplaylists");
}

PRIVATE_FUNCTION EXIT_FUNCTION
list_plt_content__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 1, "Invalid argument number");
    char buff[LKT_MESSAGE_MAX];
    safe_snprintf(buff, LKT_MESSAGE_MAX, "listplaylist %s", args->argv[0]);
    ___continuation_calls(buff);
    LOG_UNREACHABLE;
}

PRIVATE_FUNCTION EXIT_FUNCTION
search_with_cmd__(struct cmd_args *args, const char *cmd)
{
    char buff[LKT_MESSAGE_MAX];
    char regex[LKT_MESSAGE_MAX];
    int continuation = 0;
    FILE *sock       = NULL;

    FAIL_IF(lkt_get_query_type(args, regex, LKT_MESSAGE_MAX), "Query is invalid");

redo:
    sock = lkt_connect();
    write_socket(sock, "%d %s %s\n", continuation, cmd, regex);

    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);

        if (STR_NMATCH(buff, "continue:", strlen("continue:"))) {
            continuation = atoi(lkt_skip_key(buff));
            fclose(sock);
            goto redo;
        }

        HANDLE_STATUS(buff, { exit(EXIT_SUCCESS); });
        row_print(buff);
    }
}

PRIVATE_FUNCTION EXIT_FUNCTION
search_get__(struct cmd_args *args)
{
    FAIL_IF(args->argc != 1, "Invalid number of arguments");
    FAIL_IF(!strtol(args->argv[0], NULL, 0), "Invalid id");
    char buff[LKT_MESSAGE_MAX];
    char UNUSED sink;
    FILE *sock = lkt_connect();

    write_socket(sock, "find %s\n", args->argv[0]);
    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);
        HANDLE_STATUS(buff, { exit(EXIT_SUCCESS); });
        write(1, buff, strlen(buff));
    }
}

PRIVATE_FUNCTION EXIT_FUNCTION
search_plt__(struct cmd_args *args)
{
    if (args->argc == 0)
        list_plt__(args);
    if (args->argc == 1)
        list_plt_content__(args);

    char buff[LKT_MESSAGE_MAX];
    char regex[LKT_MESSAGE_MAX];
    int continuation           = 0;
    FILE *sock                 = NULL;
    struct cmd_args args_regex = CMD_ARGS_NEXT(*args);

    FAIL_IF(args->argc < 2, "Invalid number of arguments");
    FAIL_IF(lkt_get_query_type(&args_regex, regex, LKT_MESSAGE_MAX), "Query is invalid");

redo:
    sock = lkt_connect();
    write_socket(sock, "%d listplaylistinfo %s %s\n", continuation, args->argv[0], regex);

    FOR_EVER
    {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        read_socket(sock, buff, LKT_MESSAGE_MAX - 1);

        if (STR_NMATCH(buff, "continue:", strlen("continue:"))) {
            continuation = atoi(lkt_skip_key(buff));
            fclose(sock);
            goto redo;
        }

        HANDLE_STATUS(buff, { exit(EXIT_SUCCESS); });
        row_print(buff);
    }
}

#define search_with_cmd(func, cmd) /* I don't want to write always the same things */ \
    PRIVATE_FUNCTION EXIT_FUNCTION func(struct cmd_args *args)                        \
    {                                                                                 \
        search_with_cmd__(args, #cmd);                                                \
    }
// clang-format off
search_with_cmd(search_db__, search);
search_with_cmd(search_count__, count);
search_with_cmd(search_queue__, playlistfind);
// clang-format on
#undef search_with_cmd

/* Parsing stuff. */

// clang-format off
static struct cmd_opt options_queue[] = {
    { .name = "insert",     .call = queue_insert__      },
    { .name = "pos",        .call = queue_pos__         },
    { .name = "pop",        .call = queue_pop__         },
    { .name = "add",        .call = queue_add__         },
    { .name = "seek",       .call = queue_seek__        },
    { .name = "delete",     .call = queue_delete__      },
    { .name = "remove",     .call = queue_remove__      },
    { .name = "clear",      .call = queue_clear__       },
    { .name = "crop",       .call = queue_crop__        },
    { .name = "replace",    .call = queue_replace__     },
    { .name = "dump",       .call = queue_dump__        },
    { .name = "id",         .call = queue_id__          },
    { .name = "flatten",    .call = queue_flatten__     },
    { .name = "swap",       .call = queue_swap__        },
    { .name = "swapid",     .call = queue_swapid__      },
    CMD_OPT_DEFAULT(queue_list__),
};

static struct cmd_opt options_plt[] = {
    { .name = "add",        .call = plt_add__           },
    { .name = "delete",     .call = plt_delete__        },
    { .name = "destroy",    .call = plt_destroy__       },
    { .name = "create",     .call = plt_create__        },
    { .name = "list",       .call = search_plt__        },
    CMD_OPT_NULL,
};

static struct cmd_opt options_search[] = {
    { .name = "database",   .call = search_db__         },
    { .name = "get",        .call = search_get__        },
    { .name = "plt",        .call = search_plt__        },
    { .name = "count",      .call = search_count__      },
    { .name = "queue",      .call = search_queue__      },
    CMD_OPT_DEFAULT(search_db__),
};

static struct cmd_opt options_admin[] = {
    { .name = "ping",       .call = ping__              },
    { .name = "kill",       .call = kill__              },
    { .name = "restart",    .call = restart__           },
    { .name = "rescan",     .call = rescan__            },
    { .name = "update",     .call = update__            },
    { .name = "dry-update", .call = dry_update__        },
    { .name = "populate",   .call = populate__          },
    { .name = "config",     .call = config__            },
    { .name = "import",     .call = import__            },
    CMD_OPT_NULL,
};

static struct cmd_opt options_stickers[] = {
    { .name = "get",        .call = stickers_get__      },
    { .name = "set",        .call = stickers_set__      },
    { .name = "delete",     .call = stickers_delete__   },
    { .name = "create",     .call = stickers_create__   },
    { .name = "destroy",    .call = stickers_destroy__  },
    CMD_OPT_NULL,
};
// clang-format on

#define sub_command(name) /* Create sub-commands here */                           \
    PRIVATE_FUNCTION EXIT_FUNCTION name##__(struct cmd_args *args)                 \
    {                                                                              \
        FAIL_IF(!args->argc, "Invalid command, specify a sub command for " #name); \
        cmd_parse(options_##name, args->argc, args->argv);                         \
    }
sub_command(stickers);
sub_command(search);
sub_command(plt);
sub_command(admin);
#undef sub_command

PRIVATE_FUNCTION EXIT_FUNCTION
queue__(struct cmd_args *args)
{
    if (args->argc == 0)
        queue_list__(args);
    cmd_parse(options_queue, args->argc, args->argv);
}

// clang-format off
static struct cmd_opt options_[] = {
    { .name = "current",    .call = current__   },
    { .name = "play",       .call = play__      },
    { .name = "pause",      .call = pause__     },
    { .name = "unpause",    .call = unpause__   },
    { .name = "next",       .call = next__      },
    { .name = "previous",   .call = prev__      },
    { .name = "queue",      .call = queue__     },
    { .name = "shuffle",    .call = shuffle__   },
    { .name = "status",     .call = status__    },
    { .name = "stop",       .call = stop__      },
    { .name = "plt",        .call = plt__       },
    { .name = "search",     .call = search__    },
    { .name = "admin",      .call = admin__     },
    { .name = "stickers",   .call = stickers__  },
    CMD_OPT_NULL,
};
// clang-format on

/* The sigpipe function, if SIGPIPE signal is sent. */

PRIVATE_FUNCTION EXIT_FUNCTION
sigpipe__(int sig)
{
    LOG_ERROR("GENERAL", "Exit because of signal sigpipe (%d)", sig);
    exit(EXIT_FAILURE);
}

/* Handle the options switches */

#define ___DEFINE_SETTER(what)                                                            \
    PRIVATE_FUNCTION void ___USE_SETTER(what)(const char UNUSED *name, const char *value, \
                                              void UNUSED *user)                          \
    {                                                                                     \
        what = value;                                                                     \
    }
#define ___USE_SETTER(what) ___set_env_##what

// clang-format off

___DEFINE_SETTER(host);
___DEFINE_SETTER(port);
___DEFINE_SETTER(pwd);

PRIVATE_FUNCTION void
___USE_SETTER(row_format)(const char UNUSED *name, const char *value, void UNUSED *user)
{
    row_format = value;
    setup_row_printer();
}

static struct cmd_env env_[] = {
    CMD_ENV_DEFAULT("host", "localhost", ___USE_SETTER(host)),
    CMD_ENV_DEFAULT("port", "6600",      ___USE_SETTER(port)),
    CMD_ENV("pwd",                       ___USE_SETTER(pwd)),
    CMD_ENV("row_format",                ___USE_SETTER(row_format)),
    CMD_ENV_NULL,
};

// clang-format on

#undef ___USE_SETTER
#undef ___DEFINE_SETTER

struct lkt_logfile *___log_logfile = NULL;
void lkt_logfile_vfwrite(struct lkt_logfile *, const char *, va_list);
void lkt_logfile_write(struct lkt_logfile *, const char *);
void
lkt_logfile_vfwrite(UNUSED struct lkt_logfile *logfile, UNUSED const char *line, UNUSED va_list ap)
{
}
void
lkt_logfile_write(UNUSED struct lkt_logfile *logfile, UNUSED const char *line)
{
}

int
main(int argc, const char **argv)
{
    lkt_set_log_level_verbose(true);
    lkt_segv_quiet();
    lkt_install_segv_handler();
    lkt_set_log_level(LOG_LEVEL_INFO);
    cmd_set_executable_name("lkt");

    /* Enforce some locals */
    RETURN_UNLESS(setlocale(LC_ALL, ""), "Failed to set LC_ALL to UTF-8", 1);
    RETURN_UNLESS(setlocale(LC_CTYPE, ""), "Failed to set LC_CTYPE", 1);
    RETURN_UNLESS(setlocale(LC_NUMERIC, ""), "Failed to set LC_NUMERIC for mpv", 1);
    RETURN_UNLESS(STR_MATCH(nl_langinfo(CODESET), "UTF-8"),
                  "Your locale is not set to an UTF-8 one. "
                  "Consider using en_US.UTF-8 or fr_FR.UTF-8!",
                  1);

    if (signal(SIGPIPE, sigpipe__))
        LOG_ERROR("SYS", "Failed to install handler for SIGPIPE signal (you may be using php...)");

    /* Parse */
    cmd_parse_env(env_, &argc, &argv);
    cmd_apply_env(env_, NULL);
    cmd_parse(options_, argc, argv);
}
