use egui::TextBuffer;
use serde::{Deserialize, Serialize};
use std::{fmt, marker::PhantomData, str::FromStr};

#[derive(Serialize, Deserialize)]
pub struct IntegerTextBuffer<T: private::Unsigned + private::Zero<T>> {
    buffer: String,
    phantom: PhantomData<T>,
}

impl<T: private::Unsigned + private::Zero<T>> IntegerTextBuffer<T> {
    pub fn as_integer(&self) -> T
    where
        <T as FromStr>::Err: std::fmt::Display,
    {
        if self.buffer.is_empty() {
            T::ZERO
        } else {
            match self.buffer.parse::<T>() {
                Ok(int) => int,
                Err(e) => panic!("Failed to parse integer: {}", e),
            }
        }
    }

    fn remove_all_non_digit_chars(&mut self) {
        self.buffer = self.buffer.chars().filter(|c| c.is_ascii_digit()).collect();
    }
}

impl<T: private::Unsigned + private::Zero<T>> fmt::Display for IntegerTextBuffer<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.buffer)
    }
}

impl<T: private::Unsigned + private::Zero<T>> From<&str> for IntegerTextBuffer<T> {
    fn from(text: &str) -> Self {
        Self {
            buffer: text.to_string(),
            phantom: PhantomData,
        }
    }
}

impl<T: private::Unsigned + private::Zero<T>> AsRef<str> for IntegerTextBuffer<T> {
    fn as_ref(&self) -> &str {
        self.buffer.as_str()
    }
}

impl<T: private::Unsigned + private::Zero<T>> TextBuffer for IntegerTextBuffer<T> {
    fn is_mutable(&self) -> bool {
        true
    }

    fn as_str(&self) -> &str {
        self.as_ref()
    }

    fn char_range(&self, char_range: std::ops::Range<usize>) -> &str {
        assert!(char_range.start <= char_range.end);
        let start_byte = self.byte_index_from_char_index(char_range.start);
        let end_byte = self.byte_index_from_char_index(char_range.end);
        &self.as_str()[start_byte..end_byte]
    }

    fn insert_text(&mut self, text: &str, char_index: usize) -> usize {
        let ret = self.buffer.insert_text(text, char_index);
        self.remove_all_non_digit_chars();
        ret
    }

    fn delete_char_range(&mut self, char_range: std::ops::Range<usize>) {
        self.buffer.delete_char_range(char_range);
    }

    fn clear(&mut self) {
        self.delete_char_range(0..self.as_ref().len());
    }

    fn replace(&mut self, text: &str) {
        self.clear();
        self.insert_text(text, 0);
    }

    fn take(&mut self) -> String {
        let s = self.as_ref().to_owned();
        self.clear();
        s
    }
}

mod private {
    use std::str::FromStr;

    pub trait Unsigned: FromStr {}
    impl Unsigned for u8 {}
    impl Unsigned for u16 {}
    impl Unsigned for u32 {}
    impl Unsigned for u64 {}
    impl Unsigned for u128 {}
    impl Unsigned for usize {}

    pub trait Zero<T: Unsigned> {
        const ZERO: T;
    }

    impl Zero<u8> for u8 {
        const ZERO: u8 = 0u8;
    }

    impl Zero<u16> for u16 {
        const ZERO: u16 = 0u16;
    }

    impl Zero<u32> for u32 {
        const ZERO: u32 = 0u32;
    }

    impl Zero<u64> for u64 {
        const ZERO: u64 = 0u64;
    }

    impl Zero<u128> for u128 {
        const ZERO: u128 = 0u128;
    }

    impl Zero<usize> for usize {
        const ZERO: usize = 0usize;
    }
}
