//! Contains types for typed response.

use crate::types::LektorState;
use amadeus_macro::*;

/// A formated response is just a list of key/pairs. We get every line that is
/// not Ok/Ack/Continue (i.e. data lines) and split on the first ':' and trim
/// spaces from the keys and the values. The keys are always in lowercase.
#[derive(Debug)]
pub struct LektorFormatedResponse {
    content: Vec<(String, String)>,
    raw_content: Vec<String>,
}

impl LektorFormatedResponse {
    /// Pop the first key found in the response, get an error if the key is not
    /// found. If multiple keys are found, only the first found is returned.
    pub fn pop(&mut self, key: &str) -> Result<String, String> {
        match self
            .content
            .iter()
            .enumerate()
            .find_map(|(index, (what, _))| then_some!(what == key => index))
        {
            Some(index) => Ok(self.content.remove(index).1),
            None => Err(format!("no key {key} was found in formated response")),
        }
    }

    /// Pop all the entries with the said key. If no key is found then the empty
    /// vector is returned. This function can't fail.
    ///
    /// FIXME: Get ride of the clone in this function...
    pub fn pop_all(&mut self, key: &str) -> Vec<String> {
        let mut ret = Vec::new();
        self.content.retain(|(what, field)| {
            if *what == key {
                ret.push(field.clone());
                false
            } else {
                true
            }
        });
        ret
    }

    /// Get the raw content of the response.
    pub fn pop_raw(self) -> Vec<String> {
        self.raw_content
    }
}

impl IntoIterator for LektorFormatedResponse {
    type Item = (String, String);
    type IntoIter =
        <std::vec::Vec<(std::string::String, std::string::String)> as IntoIterator>::IntoIter;

    fn into_iter(self) -> Self::IntoIter {
        self.content.into_iter()
    }
}

impl TryFrom<Vec<String>> for LektorFormatedResponse {
    type Error = String;

    fn try_from(vec: Vec<String>) -> Result<Self, Self::Error> {
        let (mut content, mut raw_content) = (Vec::new(), Vec::new());
        for line in vec {
            match line.splitn(2, ':').collect::<Vec<&str>>()[..] {
                [key, value] => content.push((key.trim().to_lowercase(), value.trim().to_string())),
                _ => raw_content.push(line),
            }
        }
        Ok(Self {
            content,
            raw_content,
        })
    }
}

#[derive(Debug)]
pub enum LektorResponse {
    PlaybackStatus(LektorPlaybackStatusResponse),
    CurrentKara(LektorCurrentKaraResponse),
    PlaylistSet(LektorPlaylistSetResponse),
    EmptyResponse(LektorEmptyResponse),
    KaraSet(LektorKaraSetResponse),
}

/// A trait for typed lektor responses. Such responses must be built by
/// consuming a formated response. We also protect from implemeting this trait
/// outside of this crate.
pub trait FromLektorResponse: std::fmt::Debug + private::Sealed {
    /// Consume a formated response to produce the correctly typed response. May
    /// got an error as a string that describes the problem.
    fn from_formated(response: LektorFormatedResponse) -> Result<LektorResponse, String>
    where
        Self: Sized;
}

mod private {
    use super::*;
    pub trait Sealed {}
    impl Sealed for LektorPlaybackStatusResponse {}
    impl Sealed for LektorCurrentKaraResponse {}
    impl Sealed for LektorPlaylistSetResponse {}
    impl Sealed for LektorKaraSetResponse {}
    impl Sealed for LektorEmptyResponse {}
}

macro_rules! getter {
    ($name: ident: ref $type: ty) => {
        pub fn $name(&self) -> &$type {
            &self.$name
        }
    };

    ($name: ident: $type: ty) => {
        pub fn $name(&self) -> $type {
            self.$name
        }
    };
}

#[derive(Debug)]
pub struct LektorPlaybackStatusResponse {
    elapsed: usize,
    songid: Option<usize>,
    song: Option<usize>,
    volume: usize,
    state: LektorState,
    duration: usize,
    updating_db: usize,
    playlistlength: usize,
    random: bool,
    consume: bool,
    single: bool,
    repeat: bool,
}

#[derive(Debug)]
pub struct LektorPlaylistSetResponse {
    playlists: Vec<String>,
}

#[derive(Debug)]
pub struct LektorCurrentKaraInnerResponse {
    title: String,
    author: String,
    source: String,
    song_type: String,
    song_number: Option<usize>,
    category: String,
    language: String,
}

#[derive(Debug)]
pub struct LektorKaraSetResponse {
    karas: Vec<String>,
}

#[derive(Debug)]
pub struct LektorCurrentKaraResponse {
    content: Option<LektorCurrentKaraInnerResponse>,
}

#[derive(Debug)]
pub struct LektorEmptyResponse;

impl LektorCurrentKaraResponse {
    getter!(content: ref Option<LektorCurrentKaraInnerResponse>);

    pub fn maybe_into_inner(self) -> Option<LektorCurrentKaraInnerResponse> {
        self.content
    }

    pub fn into_inner(self) -> LektorCurrentKaraInnerResponse {
        self.content.unwrap()
    }
}

impl LektorCurrentKaraInnerResponse {
    getter!(title: ref String);
    getter!(author: ref String);
    getter!(source: ref String);
    getter!(song_type: ref String);
    getter!(song_number: Option<usize>);
    getter!(category: ref String);
    getter!(language: ref String);
}

impl LektorPlaybackStatusResponse {
    getter!(elapsed: usize);
    getter!(songid: Option<usize>);
    getter!(song: Option<usize>);
    getter!(volume: usize);
    getter!(state: LektorState);
    getter!(duration: usize);
    getter!(updating_db: usize);
    getter!(playlistlength: usize);
    getter!(random: bool);
    getter!(consume: bool);
    getter!(single: bool);
    getter!(repeat: bool);

    pub fn verify(&self) -> bool {
        self.elapsed() <= self.duration()
    }
}

impl LektorKaraSetResponse {
    pub fn iter(&self) -> &[String] {
        &self.karas[..]
    }
}

impl LektorPlaylistSetResponse {
    pub fn iter(&self) -> &[String] {
        &self.playlists[..]
    }
}

impl IntoIterator for LektorKaraSetResponse {
    type Item = String;
    type IntoIter = <std::vec::Vec<std::string::String> as IntoIterator>::IntoIter;
    fn into_iter(self) -> Self::IntoIter {
        self.karas.into_iter()
    }
}

impl IntoIterator for LektorPlaylistSetResponse {
    type Item = String;
    type IntoIter = <std::vec::Vec<std::string::String> as IntoIterator>::IntoIter;
    fn into_iter(self) -> Self::IntoIter {
        self.playlists.into_iter()
    }
}

impl LektorCurrentKaraInnerResponse {
    /// If the response is partial we might want to return none (if we are not
    /// playing anything for example...)
    pub(self) fn is_partial(&self) -> bool {
        self.title.is_empty()
            || self.author.is_empty()
            || self.source.is_empty()
            || self.song_type.is_empty()
            || self.category.is_empty()
            || self.language.is_empty()
    }
}

impl FromLektorResponse for LektorEmptyResponse {
    fn from_formated(_: LektorFormatedResponse) -> Result<LektorResponse, String> {
        Ok(LektorResponse::EmptyResponse(Self {}))
    }
}

impl FromLektorResponse for LektorPlaylistSetResponse {
    fn from_formated(mut response: LektorFormatedResponse) -> Result<LektorResponse, String> {
        Ok(LektorResponse::PlaylistSet(Self {
            playlists: response.pop_all("name"),
        }))
    }
}

impl FromLektorResponse for LektorKaraSetResponse {
    fn from_formated(response: LektorFormatedResponse) -> Result<LektorResponse, String>
    where
        Self: Sized,
    {
        Ok(LektorResponse::KaraSet(Self {
            karas: response.pop_raw(),
        }))
    }
}

impl FromLektorResponse for LektorPlaybackStatusResponse {
    fn from_formated(mut response: LektorFormatedResponse) -> Result<LektorResponse, String> {
        let mut ret = Self {
            elapsed: response.pop("elapsed")?.parse::<usize>().unwrap_or(0),
            songid: match response.pop("songid")?.parse::<isize>() {
                Ok(x) if x <= 0 => None,
                Ok(x) => Some(x as usize),
                Err(_) => None,
            },
            song: match response.pop("song")?.parse::<usize>() {
                Ok(x) => Some(x),
                Err(_) => None,
            },
            volume: response
                .pop("volume")?
                .parse::<usize>()
                .unwrap()
                .clamp(0, 100),
            duration: response.pop("duration")?.parse::<usize>().unwrap(),
            updating_db: response.pop("updating_db")?.parse::<usize>().unwrap(),
            playlistlength: response.pop("playlistlength")?.parse::<usize>().unwrap(),
            random: response.pop("random")?.parse::<usize>().unwrap() != 0,
            consume: response.pop("consume")?.parse::<usize>().unwrap() != 0,
            single: response.pop("single")?.parse::<usize>().unwrap() != 0,
            repeat: response.pop("repeat")?.parse::<usize>().unwrap() != 0,
            state: LektorState::Stopped,
        };
        ret.state = match &response.pop("state")?[..] {
            "play" => LektorState::Play(ret.songid.unwrap()),
            "pause" => LektorState::Pause(ret.songid.unwrap()),
            _ => LektorState::Stopped,
        };
        Ok(LektorResponse::PlaybackStatus(ret))
    }
}

impl FromLektorResponse for LektorCurrentKaraResponse {
    fn from_formated(mut response: LektorFormatedResponse) -> Result<LektorResponse, String> {
        let song_type_number = response.pop("type")?;
        let (song_type, song_number) = match song_type_number.find(char::is_numeric) {
            Some(index) => (
                (&song_type_number[..index]).to_owned(),
                match (&song_type_number[index..]).parse::<usize>() {
                    Ok(x) => Some(x),
                    Err(_) => None,
                },
            ),
            None => panic!("Oupsy"),
        };

        let inner = LektorCurrentKaraInnerResponse {
            title: response.pop("title")?,
            author: response.pop("author")?,
            source: response.pop("source")?,
            category: response.pop("category")?,
            language: response.pop("language")?,
            song_type,
            song_number,
        };

        Ok(LektorResponse::CurrentKara(Self {
            content: then_some!(!inner.is_partial() => inner),
        }))
    }
}
